/*----- PROTECTED REGION ID(FastPS.cpp) ENABLED START -----*/
//=============================================================================
//
// file :        FastPS.cpp
//
// description : C++ source for the FastPS class and its commands.
//               The class is derived from Device. It represents the
//               CORBA servant object which will be accessed from the
//               network. All commands which can be executed on the
//               FastPS are implemented in this file.
//
// project :     CaenElsFastPS
//
// This file is part of Tango device class.
// 
// Tango is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Tango is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Tango.  If not, see <http://www.gnu.org/licenses/>.
// 
//
// Elettra-Sincrotrone Trieste S.C.p.A. di interesse nazionale
// Strada Statale 14 - km 163,5 in AREA Science Park
// 34149 Basovizza, Trieste ITALY
//
//=============================================================================
//                This file is generated by POGO
//        (Program Obviously used to Generate tango Object)
//=============================================================================


#include <FastPS.h>
#include <FastPSClass.h>

/*----- PROTECTED REGION END -----*/	//	FastPS.cpp

/**
 *  FastPS class description:
 *    Caenels Fast Powersupply tango server
 */

//================================================================
//  The following table gives the correspondence
//  between command and method names.
//
//  Command name   |  Method name
//================================================================
//  State          |  Inherited (no method)
//  Status         |  Inherited (no method)
//  On             |  on
//  Off            |  off
//  StartWaveform  |  start_waveform
//  Reset          |  reset
//  HwReset        |  hw_reset
//  StopWaveform   |  stop_waveform
//================================================================

//================================================================
//  Attributes managed are:
//================================================================
//  CurrentControl           |  Tango::DevBoolean	Scalar
//  VoltageControl           |  Tango::DevBoolean	Scalar
//  Voltage                  |  Tango::DevDouble	Scalar
//  Current                  |  Tango::DevDouble	Scalar
//  CurrentSet               |  Tango::DevDouble	Scalar
//  WaveformExternalTrigger  |  Tango::DevBoolean	Scalar
//  VoltageDCLink            |  Tango::DevDouble	Scalar
//  MosfetTemperature        |  Tango::DevDouble	Scalar
//  AnalogControl            |  Tango::DevBoolean	Scalar
//  WaveformTimeScale        |  Tango::DevDouble	Scalar
//  WaveformPeriods          |  Tango::DevLong	Scalar
//  Waveform                 |  Tango::DevDouble	Spectrum  ( max = 500000)
//  HwWaveform               |  Tango::DevDouble	Spectrum  ( max = 500000)
//================================================================

namespace FastPS_ns
{
/*----- PROTECTED REGION ID(FastPS::namespace_starting) ENABLED START -----*/

//	static initializations

/*----- PROTECTED REGION END -----*/	//	FastPS::namespace_starting

//--------------------------------------------------------
/**
 *	Method      : FastPS::FastPS()
 *	Description : Constructors for a Tango device
 *                implementing the classFastPS
 */
//--------------------------------------------------------
FastPS::FastPS(Tango::DeviceClass *cl, string &s)
 : TANGO_BASE_CLASS(cl, s.c_str())
{
	/*----- PROTECTED REGION ID(FastPS::constructor_1) ENABLED START -----*/
	init_device();
	
	/*----- PROTECTED REGION END -----*/	//	FastPS::constructor_1
}
//--------------------------------------------------------
FastPS::FastPS(Tango::DeviceClass *cl, const char *s)
 : TANGO_BASE_CLASS(cl, s)
{
	/*----- PROTECTED REGION ID(FastPS::constructor_2) ENABLED START -----*/
	init_device();
	
	/*----- PROTECTED REGION END -----*/	//	FastPS::constructor_2
}
//--------------------------------------------------------
FastPS::FastPS(Tango::DeviceClass *cl, const char *s, const char *d)
 : TANGO_BASE_CLASS(cl, s, d)
{
	/*----- PROTECTED REGION ID(FastPS::constructor_3) ENABLED START -----*/
	init_device();
	
	/*----- PROTECTED REGION END -----*/	//	FastPS::constructor_3
}

//--------------------------------------------------------
/**
 *	Method      : FastPS::delete_device()
 *	Description : will be called at device destruction or at init command
 */
//--------------------------------------------------------
void FastPS::delete_device()
{
	DEBUG_STREAM << "FastPS::delete_device() " << device_name << endl;
	/*----- PROTECTED REGION ID(FastPS::delete_device) ENABLED START -----*/
	
	//	Delete device allocated objects
	if (sock)
		close(sock);

	if (mutex)
		delete mutex;

	
	/*----- PROTECTED REGION END -----*/	//	FastPS::delete_device
	delete[] attr_CurrentControl_read;
	delete[] attr_VoltageControl_read;
	delete[] attr_Voltage_read;
	delete[] attr_Current_read;
	delete[] attr_CurrentSet_read;
	delete[] attr_WaveformExternalTrigger_read;
	delete[] attr_VoltageDCLink_read;
	delete[] attr_MosfetTemperature_read;
	delete[] attr_AnalogControl_read;
	delete[] attr_WaveformTimeScale_read;
	delete[] attr_WaveformPeriods_read;
	delete[] attr_Waveform_read;
	delete[] attr_HwWaveform_read;
}

//--------------------------------------------------------
/**
 *	Method      : FastPS::init_device()
 *	Description : will be called at device initialization.
 */
//--------------------------------------------------------
void FastPS::init_device()
{
	DEBUG_STREAM << "FastPS::init_device() create device " << device_name << endl;
	/*----- PROTECTED REGION ID(FastPS::init_device_before) ENABLED START -----*/
	
	//	Initialization before get_device_property() call
	
	/*----- PROTECTED REGION END -----*/	//	FastPS::init_device_before
	

	//	Get the device properties from database
	get_device_property();
	
	attr_CurrentControl_read = new Tango::DevBoolean[1];
	attr_VoltageControl_read = new Tango::DevBoolean[1];
	attr_Voltage_read = new Tango::DevDouble[1];
	attr_Current_read = new Tango::DevDouble[1];
	attr_CurrentSet_read = new Tango::DevDouble[1];
	attr_WaveformExternalTrigger_read = new Tango::DevBoolean[1];
	attr_VoltageDCLink_read = new Tango::DevDouble[1];
	attr_MosfetTemperature_read = new Tango::DevDouble[1];
	attr_AnalogControl_read = new Tango::DevBoolean[1];
	attr_WaveformTimeScale_read = new Tango::DevDouble[1];
	attr_WaveformPeriods_read = new Tango::DevLong[1];
	attr_Waveform_read = new Tango::DevDouble[500000];
	attr_HwWaveform_read = new Tango::DevDouble[500000];
	/*----- PROTECTED REGION ID(FastPS::init_device) ENABLED START -----*/
	
	//	Initialize device
	mutex = 0;
	mutex = new omni_mutex();

	sock_connect();

	error_mapping.insert(std::make_pair(1, "Unknown command"));
	error_mapping.insert(std::make_pair(2, "Unknown Parameter"));
	error_mapping.insert(std::make_pair(3, "Index out of range"));
	error_mapping.insert(std::make_pair(4, "Not Enough Arguments"));
	error_mapping.insert(std::make_pair(5, "Privilege Level Requirement not met"));
	error_mapping.insert(std::make_pair(6, "Save Error on device"));
	error_mapping.insert(std::make_pair(7, "Invalid password"));
	error_mapping.insert(std::make_pair(8, "Module in fault"));
	error_mapping.insert(std::make_pair(9, "Module already ON"));
	error_mapping.insert(std::make_pair(10, "Setpoint is out of hardware limits"));
	error_mapping.insert(std::make_pair(11, "Setpoint is out of software limits"));
	error_mapping.insert(std::make_pair(12, "Setpoint is not a number"));
	error_mapping.insert(std::make_pair(13, "Module is OFF"));
	error_mapping.insert(std::make_pair(14, "Slew Rate out of limits"));
	error_mapping.insert(std::make_pair(15, "Device is set in local mode"));
	error_mapping.insert(std::make_pair(16, "Module is not in waveform mode"));
	error_mapping.insert(std::make_pair(17, "Module is in waveform mode"));
	error_mapping.insert(std::make_pair(18, "Device is set in remote mode"));
	error_mapping.insert(std::make_pair(19, "Module is already in the selected loop mode"));
	error_mapping.insert(std::make_pair(20, "Module is not in the selected loop mode"));
	error_mapping.insert(std::make_pair(21, "Module is not in normal update mode Remote Control Manual"));
	error_mapping.insert(std::make_pair(22, "Float mode is already selected"));
	error_mapping.insert(std::make_pair(23, "Unknown sub-command for SFP communication"));
	error_mapping.insert(std::make_pair(24, "Unknown feature or feature not available (AIN,TRIG)"));
	error_mapping.insert(std::make_pair(25, "Parallel Fault"));
	error_mapping.insert(std::make_pair(26, "Waveform error"));
	error_mapping.insert(std::make_pair(27, "Cannot open the required file"));
	error_mapping.insert(std::make_pair(28, "Module is currently inverting the polarity"));
	error_mapping.insert(std::make_pair(29, "Cannot write waveform data"));
	error_mapping.insert(std::make_pair(30, "Polarity switch not allowed"));
	error_mapping.insert(std::make_pair(31, "Cannot set options for socket used by oscilloscope"));
	error_mapping.insert(std::make_pair(32, "Cannot change settings while in parallel slave mode"));
	error_mapping.insert(std::make_pair(33, "MASTER and SLAVES have different FW version"));
	error_mapping.insert(std::make_pair(34, "MASTER and SLAVES are different models"));
	error_mapping.insert(std::make_pair(35, "MASTER and SLAVES have different ratings"));
	error_mapping.insert(std::make_pair(36, "The required feature is not available"));
	error_mapping.insert(std::make_pair(37, "UDP buffer overflow"));
	error_mapping.insert(std::make_pair(38, "Module is in “Wait for OFF”"));
	error_mapping.insert(std::make_pair(39, "This field is read only"));
	error_mapping.insert(std::make_pair(40, "Cannot parse input name for debug field"));
	error_mapping.insert(std::make_pair(41, "Cannot parse input value for debug field"));
	error_mapping.insert(std::make_pair(42, "Cannot parse type for debug field"));
	error_mapping.insert(std::make_pair(43, "DHCP is enabled"));
	error_mapping.insert(std::make_pair(44, "This command is disabled"));
	error_mapping.insert(std::make_pair(99, "Unknown err"));

	wave_size = FASTPS_MIN_WAVE;
	for (int i = 0; i < FASTPS_MAX_WAVE; i++) {
		attr_Waveform_read[i] = 0;
	}
	*attr_WaveformTimeScale_read = 1;

	try {
		Tango::MultiAttribute *attr_list = get_device_attr();    
		attr_set_current = 0;
		attr_set_current = &(attr_list->get_w_attr_by_name("Current"));
	}
	catch(...) {
		INFO_STREAM << "A2605::init_device(): ERROR GETTING ATTRIBUTE" << endl;
	} 
	
	/*----- PROTECTED REGION END -----*/	//	FastPS::init_device
}

//--------------------------------------------------------
/**
 *	Method      : FastPS::get_device_property()
 *	Description : Read database to initialize property data members.
 */
//--------------------------------------------------------
void FastPS::get_device_property()
{
	/*----- PROTECTED REGION ID(FastPS::get_device_property_before) ENABLED START -----*/
	
	//	Initialize property data members
	
	/*----- PROTECTED REGION END -----*/	//	FastPS::get_device_property_before


	//	Read device properties from database.
	Tango::DbData	dev_prop;
	dev_prop.push_back(Tango::DbDatum("IpAddress"));
	dev_prop.push_back(Tango::DbDatum("TcpPort"));
	dev_prop.push_back(Tango::DbDatum("TxRxSleep"));

	//	is there at least one property to be read ?
	if (dev_prop.size()>0)
	{
		//	Call database and extract values
		if (Tango::Util::instance()->_UseDb==true)
			get_db_device()->get_property(dev_prop);
	
		//	get instance on FastPSClass to get class property
		Tango::DbDatum	def_prop, cl_prop;
		FastPSClass	*ds_class =
			(static_cast<FastPSClass *>(get_device_class()));
		int	i = -1;

		//	Try to initialize IpAddress from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  ipAddress;
		else {
			//	Try to initialize IpAddress from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  ipAddress;
		}
		//	And try to extract IpAddress value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  ipAddress;

		//	Try to initialize TcpPort from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  tcpPort;
		else {
			//	Try to initialize TcpPort from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  tcpPort;
		}
		//	And try to extract TcpPort value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  tcpPort;

		//	Try to initialize TxRxSleep from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  txRxSleep;
		else {
			//	Try to initialize TxRxSleep from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  txRxSleep;
		}
		//	And try to extract TxRxSleep value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  txRxSleep;

	}

	/*----- PROTECTED REGION ID(FastPS::get_device_property_after) ENABLED START -----*/
	
	//	Check device property data members init
	
	/*----- PROTECTED REGION END -----*/	//	FastPS::get_device_property_after
}

//--------------------------------------------------------
/**
 *	Method      : FastPS::always_executed_hook()
 *	Description : method always executed before any command is executed
 */
//--------------------------------------------------------
void FastPS::always_executed_hook()
{
	DEBUG_STREAM << "FastPS::always_executed_hook()  " << device_name << endl;
	/*----- PROTECTED REGION ID(FastPS::always_executed_hook) ENABLED START -----*/
	
	//	code always executed before all requests


	struct timeval curr_state_time;
	stringstream s;

	try {
		set_DevState(); 
	}
	catch(...){}


	gettimeofday(&curr_state_time, NULL);
	if (abs(curr_state_time.tv_sec - socket_state_time.tv_sec) > FASTPS_RECONNECT_MAXERR) {
		set_state(Tango::FAULT);
		s << "Reconnection error" << endl;
		if (sock)
			close(sock);
		sock_connect();
		gettimeofday(&socket_state_time,NULL);	
	}


	
	/*----- PROTECTED REGION END -----*/	//	FastPS::always_executed_hook
}

//--------------------------------------------------------
/**
 *	Method      : FastPS::read_attr_hardware()
 *	Description : Hardware acquisition for attributes
 */
//--------------------------------------------------------
void FastPS::read_attr_hardware(TANGO_UNUSED(vector<long> &attr_list))
{
	DEBUG_STREAM << "FastPS::read_attr_hardware(vector<long> &attr_list) entering... " << endl;
	/*----- PROTECTED REGION ID(FastPS::read_attr_hardware) ENABLED START -----*/
	
	//	Add your own code
	
	/*----- PROTECTED REGION END -----*/	//	FastPS::read_attr_hardware
}
//--------------------------------------------------------
/**
 *	Method      : FastPS::write_attr_hardware()
 *	Description : Hardware writing for attributes
 */
//--------------------------------------------------------
void FastPS::write_attr_hardware(TANGO_UNUSED(vector<long> &attr_list))
{
	DEBUG_STREAM << "FastPS::write_attr_hardware(vector<long> &attr_list) entering... " << endl;
	/*----- PROTECTED REGION ID(FastPS::write_attr_hardware) ENABLED START -----*/
	
	//	Add your own code
	
	/*----- PROTECTED REGION END -----*/	//	FastPS::write_attr_hardware
}

//--------------------------------------------------------
/**
 *	Read attribute CurrentControl related method
 *	Description: Enable current control mode
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void FastPS::read_CurrentControl(Tango::Attribute &attr)
{
	DEBUG_STREAM << "FastPS::read_CurrentControl(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(FastPS::read_CurrentControl) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_CurrentControl_read);
	
	/*----- PROTECTED REGION END -----*/	//	FastPS::read_CurrentControl
}
//--------------------------------------------------------
/**
 *	Write attribute CurrentControl related method
 *	Description: Enable current control mode
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void FastPS::write_CurrentControl(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "FastPS::write_CurrentControl(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(FastPS::write_CurrentControl) ENABLED START -----*/
	char strin[SHORT_LEN],strout[SHORT_LEN];
	memset(strin,0,SHORT_LEN);memset(strout,0,SHORT_LEN);

	if (w_val == true)
		sprintf(strin,"LOOP:I");
	else
		sprintf(strout,"LOOP:V");

	write_cmd(strin,strout,txRxSleep);	
		

	/*----- PROTECTED REGION END -----*/	//	FastPS::write_CurrentControl
}
//--------------------------------------------------------
/**
 *	Read attribute VoltageControl related method
 *	Description: Enable voltage control mode
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void FastPS::read_VoltageControl(Tango::Attribute &attr)
{
	DEBUG_STREAM << "FastPS::read_VoltageControl(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(FastPS::read_VoltageControl) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_VoltageControl_read);
	
	/*----- PROTECTED REGION END -----*/	//	FastPS::read_VoltageControl
}
//--------------------------------------------------------
/**
 *	Write attribute VoltageControl related method
 *	Description: Enable voltage control mode
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void FastPS::write_VoltageControl(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "FastPS::write_VoltageControl(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(FastPS::write_VoltageControl) ENABLED START -----*/
	char strin[SHORT_LEN],strout[SHORT_LEN];
	memset(strin,0,SHORT_LEN);memset(strout,0,SHORT_LEN);

	if (w_val == true)
		sprintf(strin,"LOOP:V");
	else
		sprintf(strout,"LOOP:I");

	write_cmd(strin,strout,txRxSleep);	
	
	/*----- PROTECTED REGION END -----*/	//	FastPS::write_VoltageControl
}
//--------------------------------------------------------
/**
 *	Read attribute Voltage related method
 *	Description: Read voltage
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void FastPS::read_Voltage(Tango::Attribute &attr)
{
	DEBUG_STREAM << "FastPS::read_Voltage(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(FastPS::read_Voltage) ENABLED START -----*/
	//	Set the attribute value
	char strin[SHORT_LEN],strout[SHORT_LEN];
	memset(strin,0,SHORT_LEN);memset(strout,0,SHORT_LEN);

	sprintf(strin,"MRV:?");
	write_cmd(strin,strout,txRxSleep);	
	//#MRI:2.3121
	*attr_Voltage_read = atof(strout+strlen(strin)-1);

	attr.set_value(attr_Voltage_read);
	
	/*----- PROTECTED REGION END -----*/	//	FastPS::read_Voltage
}
//--------------------------------------------------------
/**
 *	Read attribute Current related method
 *	Description: Read current
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void FastPS::read_Current(Tango::Attribute &attr)
{
	DEBUG_STREAM << "FastPS::read_Current(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(FastPS::read_Current) ENABLED START -----*/
	//	Set the attribute value
	char strin[SHORT_LEN],strout[SHORT_LEN];
	memset(strin,0,SHORT_LEN);memset(strout,0,SHORT_LEN);

	sprintf(strin,"MRI:?");
	write_cmd(strin,strout,txRxSleep);	
	//#MRI:2.3121
	*attr_Current_read = atof(strout+strlen(strin)-1);
	attr_set_current->set_write_value(*attr_Current_read);

	attr.set_value(attr_Current_read);
	
	/*----- PROTECTED REGION END -----*/	//	FastPS::read_Current
}
//--------------------------------------------------------
/**
 *	Write attribute Current related method
 *	Description: Read current
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void FastPS::write_Current(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "FastPS::write_Current(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevDouble	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(FastPS::write_Current) ENABLED START -----*/
	char strin[SHORT_LEN],strout[SHORT_LEN];
	memset(strin,0,SHORT_LEN);memset(strout,0,SHORT_LEN);

	sprintf(strin,"MWI:%.4f",w_val);
	write_cmd(strin,strout,txRxSleep);	
	
	/*----- PROTECTED REGION END -----*/	//	FastPS::write_Current
}
//--------------------------------------------------------
/**
 *	Read attribute CurrentSet related method
 *	Description: Read current last set
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void FastPS::read_CurrentSet(Tango::Attribute &attr)
{
	DEBUG_STREAM << "FastPS::read_CurrentSet(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(FastPS::read_CurrentSet) ENABLED START -----*/
	//	Set the attribute value
	char strin[SHORT_LEN],strout[SHORT_LEN];
	memset(strin,0,SHORT_LEN);memset(strout,0,SHORT_LEN);

	sprintf(strin,"MRI:?");
	write_cmd(strin,strout,txRxSleep);	
	*attr_CurrentSet_read = atof(strout+strlen(strin)-1);

	attr.set_value(attr_CurrentSet_read);
	
	/*----- PROTECTED REGION END -----*/	//	FastPS::read_CurrentSet
}
//--------------------------------------------------------
/**
 *	Read attribute WaveformExternalTrigger related method
 *	Description: Enable external waveform trigger
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void FastPS::read_WaveformExternalTrigger(Tango::Attribute &attr)
{
	DEBUG_STREAM << "FastPS::read_WaveformExternalTrigger(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(FastPS::read_WaveformExternalTrigger) ENABLED START -----*/
	//	Set the attribute value
	char strin[SHORT_LEN],strout[SHORT_LEN];
	memset(strin,0,SHORT_LEN);memset(strout,0,SHORT_LEN);

	sprintf(strin,"TRIG:?");
	write_cmd(strin,strout,txRxSleep);	
	//#MRI:2.3121
	if (strncmp(strout,"#TRIG:OFF",9) == 0)
		*attr_WaveformExternalTrigger_read = false;
	else
		*attr_WaveformExternalTrigger_read = true;

	attr.set_value(attr_WaveformExternalTrigger_read);
	
	/*----- PROTECTED REGION END -----*/	//	FastPS::read_WaveformExternalTrigger
}
//--------------------------------------------------------
/**
 *	Write attribute WaveformExternalTrigger related method
 *	Description: Enable external waveform trigger
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void FastPS::write_WaveformExternalTrigger(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "FastPS::write_WaveformExternalTrigger(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(FastPS::write_WaveformExternalTrigger) ENABLED START -----*/
	char strin[SHORT_LEN],strout[SHORT_LEN];
	memset(strin,0,SHORT_LEN);memset(strout,0,SHORT_LEN);

	if (w_val == true) 
		sprintf(strin,"TRIG:POS");
	else
		sprintf(strin,"TRIG:OFF");

	write_cmd(strin,strout,txRxSleep);	
	


	/*----- PROTECTED REGION END -----*/	//	FastPS::write_WaveformExternalTrigger
}
//--------------------------------------------------------
/**
 *	Read attribute VoltageDCLink related method
 *	Description: Read voltage DC link
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void FastPS::read_VoltageDCLink(Tango::Attribute &attr)
{
	DEBUG_STREAM << "FastPS::read_VoltageDCLink(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(FastPS::read_VoltageDCLink) ENABLED START -----*/
	//	Set the attribute value
	char strin[SHORT_LEN],strout[SHORT_LEN];
	memset(strin,0,SHORT_LEN);memset(strout,0,SHORT_LEN);

	sprintf(strin,"MRP:?");
	write_cmd(strin,strout,txRxSleep);
	*attr_VoltageDCLink_read = atof(strout+strlen(strin)-1);	

	attr.set_value(attr_VoltageDCLink_read);
	
	/*----- PROTECTED REGION END -----*/	//	FastPS::read_VoltageDCLink
}
//--------------------------------------------------------
/**
 *	Read attribute MosfetTemperature related method
 *	Description: Read mosfet heatsink temperature
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void FastPS::read_MosfetTemperature(Tango::Attribute &attr)
{
	DEBUG_STREAM << "FastPS::read_MosfetTemperature(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(FastPS::read_MosfetTemperature) ENABLED START -----*/
	//	Set the attribute value
	char strin[SHORT_LEN],strout[SHORT_LEN];
	memset(strin,0,SHORT_LEN);memset(strout,0,SHORT_LEN);

	sprintf(strin,"MRT:?");
	write_cmd(strin,strout,txRxSleep);	
	*attr_MosfetTemperature_read = atof(strout+strlen(strin)-1);	

	attr.set_value(attr_MosfetTemperature_read);
	
	/*----- PROTECTED REGION END -----*/	//	FastPS::read_MosfetTemperature
}
//--------------------------------------------------------
/**
 *	Read attribute AnalogControl related method
 *	Description: Enable external analog control
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void FastPS::read_AnalogControl(Tango::Attribute &attr)
{
	DEBUG_STREAM << "FastPS::read_AnalogControl(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(FastPS::read_AnalogControl) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_AnalogControl_read);
	
	/*----- PROTECTED REGION END -----*/	//	FastPS::read_AnalogControl
}
//--------------------------------------------------------
/**
 *	Write attribute AnalogControl related method
 *	Description: Enable external analog control
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void FastPS::write_AnalogControl(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "FastPS::write_AnalogControl(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(FastPS::write_AnalogControl) ENABLED START -----*/
	char strin[SHORT_LEN],strout[SHORT_LEN];
	memset(strin,0,SHORT_LEN);memset(strout,0,SHORT_LEN);

	if (w_val == true) 
		sprintf(strin,"UPMODE:ANALOG");
	else 	
		sprintf(strin,"UPMODE:NORMAL");
	
	write_cmd(strin,strout,txRxSleep);	
	
	/*----- PROTECTED REGION END -----*/	//	FastPS::write_AnalogControl
}
//--------------------------------------------------------
/**
 *	Read attribute WaveformTimeScale related method
 *	Description: Set the timescale multiplication factor.
 *               By defaul tone waveform sample corresponds to 10us.
 *               This time (10us) is multiplicated by WaveformTimeScale factor.
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void FastPS::read_WaveformTimeScale(Tango::Attribute &attr)
{
	DEBUG_STREAM << "FastPS::read_WaveformTimeScale(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(FastPS::read_WaveformTimeScale) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_WaveformTimeScale_read);
	
	/*----- PROTECTED REGION END -----*/	//	FastPS::read_WaveformTimeScale
}
//--------------------------------------------------------
/**
 *	Write attribute WaveformTimeScale related method
 *	Description: Set the timescale multiplication factor.
 *               By defaul tone waveform sample corresponds to 10us.
 *               This time (10us) is multiplicated by WaveformTimeScale factor.
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void FastPS::write_WaveformTimeScale(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "FastPS::write_WaveformTimeScale(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevDouble	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(FastPS::write_WaveformTimeScale) ENABLED START -----*/
	if (*attr_WaveformTimeScale_read != w_val)
		write_waveform(attr_Waveform_read,wave_size,w_val);

	*attr_WaveformTimeScale_read = w_val;
	


	/*----- PROTECTED REGION END -----*/	//	FastPS::write_WaveformTimeScale
}
//--------------------------------------------------------
/**
 *	Read attribute WaveformPeriods related method
 *	Description: Set waveform periodns. 0 means infinite loop
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void FastPS::read_WaveformPeriods(Tango::Attribute &attr)
{
	DEBUG_STREAM << "FastPS::read_WaveformPeriods(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(FastPS::read_WaveformPeriods) ENABLED START -----*/
	//	Set the attribute value
	char strin[SHORT_LEN],strout[SHORT_LEN];
	memset(strin,0,SHORT_LEN);memset(strout,0,SHORT_LEN);

	sprintf(strin,"WAVEFORM:N_PERIODS:?");
	write_cmd(strin,strout,txRxSleep);
	*attr_WaveformPeriods_read = atof(strout+strlen(strin)-1);	

	attr.set_value(attr_WaveformPeriods_read);
	
	/*----- PROTECTED REGION END -----*/	//	FastPS::read_WaveformPeriods
}
//--------------------------------------------------------
/**
 *	Write attribute WaveformPeriods related method
 *	Description: Set waveform periodns. 0 means infinite loop
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void FastPS::write_WaveformPeriods(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "FastPS::write_WaveformPeriods(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevLong	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(FastPS::write_WaveformPeriods) ENABLED START -----*/
	char strin[SHORT_LEN],strout[SHORT_LEN];
	memset(strin,0,SHORT_LEN);memset(strout,0,SHORT_LEN);

	sprintf(strin,"WAVEFORM:N_PERIODS:%d",(int)w_val);
	write_cmd(strin,strout,txRxSleep);		
	
	/*----- PROTECTED REGION END -----*/	//	FastPS::write_WaveformPeriods
}
//--------------------------------------------------------
/**
 *	Read attribute Waveform related method
 *	Description: Read/write a waveform (abs values) mitigated by TimeScale factor
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 500000
 */
//--------------------------------------------------------
void FastPS::read_Waveform(Tango::Attribute &attr)
{
	DEBUG_STREAM << "FastPS::read_Waveform(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(FastPS::read_Waveform) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_Waveform_read, wave_size);
	
	/*----- PROTECTED REGION END -----*/	//	FastPS::read_Waveform
}
//--------------------------------------------------------
/**
 *	Write attribute Waveform related method
 *	Description: Read/write a waveform (abs values) mitigated by TimeScale factor
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 500000
 */
//--------------------------------------------------------
void FastPS::write_Waveform(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "FastPS::write_Waveform(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve number of write values
	int	w_length = attr.get_write_value_length();

	//	Retrieve pointer on write values (Do not delete !)
	const Tango::DevDouble	*w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(FastPS::write_Waveform) ENABLED START -----*/
	write_waveform((double *) w_val, w_length, *attr_WaveformTimeScale_read);

	for (int i = 0; i < w_length; i++) {
		attr_Waveform_read[i] = w_val[i];
	}
	wave_size = w_length;
	
	/*----- PROTECTED REGION END -----*/	//	FastPS::write_Waveform
}
//--------------------------------------------------------
/**
 *	Read attribute HwWaveform related method
 *	Description: Read/write a waveform directly from controller (no TimeScale postprocessing)
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 500000
 */
//--------------------------------------------------------
void FastPS::read_HwWaveform(Tango::Attribute &attr)
{
	DEBUG_STREAM << "FastPS::read_HwWaveform(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(FastPS::read_HwWaveform) ENABLED START -----*/
	//	Set the attribute value

	read_waveform(attr_HwWaveform_read, &hw_wave_size);
	attr.set_value(attr_HwWaveform_read, hw_wave_size);
	
	/*----- PROTECTED REGION END -----*/	//	FastPS::read_HwWaveform
}
//--------------------------------------------------------
/**
 *	Write attribute HwWaveform related method
 *	Description: Read/write a waveform directly from controller (no TimeScale postprocessing)
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 500000
 */
//--------------------------------------------------------
void FastPS::write_HwWaveform(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "FastPS::write_HwWaveform(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve number of write values
	int	w_length = attr.get_write_value_length();

	//	Retrieve pointer on write values (Do not delete !)
	const Tango::DevDouble	*w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(FastPS::write_HwWaveform) ENABLED START -----*/
	write_waveform((double *) w_val, w_length, 1);
	
	/*----- PROTECTED REGION END -----*/	//	FastPS::write_HwWaveform
}

//--------------------------------------------------------
/**
 *	Method      : FastPS::add_dynamic_attributes()
 *	Description : Create the dynamic attributes if any
 *                for specified device.
 */
//--------------------------------------------------------
void FastPS::add_dynamic_attributes()
{
	/*----- PROTECTED REGION ID(FastPS::add_dynamic_attributes) ENABLED START -----*/
	
	//	Add your own code to create and add dynamic attributes if any
	
	/*----- PROTECTED REGION END -----*/	//	FastPS::add_dynamic_attributes
}

//--------------------------------------------------------
/**
 *	Command On related method
 *	Description: Switch power supply ON
 *
 */
//--------------------------------------------------------
void FastPS::on()
{
	DEBUG_STREAM << "FastPS::On()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(FastPS::on) ENABLED START -----*/
	
	//	Add your own code
	char strin[SHORT_LEN],strout[SHORT_LEN];
	memset(strin,0,SHORT_LEN);memset(strout,0,SHORT_LEN);

	sprintf(strin,"MON");
	write_cmd(strin,strout,txRxSleep);
	
	/*----- PROTECTED REGION END -----*/	//	FastPS::on
}
//--------------------------------------------------------
/**
 *	Command Off related method
 *	Description: Switch power supply OFF
 *
 */
//--------------------------------------------------------
void FastPS::off()
{
	DEBUG_STREAM << "FastPS::Off()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(FastPS::off) ENABLED START -----*/
	
	//	Add your own code
	char strin[SHORT_LEN],strout[SHORT_LEN];
	memset(strin,0,SHORT_LEN);memset(strout,0,SHORT_LEN);

	sprintf(strin,"MOFF");
	write_cmd(strin,strout,txRxSleep);
	
	/*----- PROTECTED REGION END -----*/	//	FastPS::off
}
//--------------------------------------------------------
/**
 *	Command StartWaveform related method
 *	Description: Start Waveform
 *
 */
//--------------------------------------------------------
void FastPS::start_waveform()
{
	DEBUG_STREAM << "FastPS::StartWaveform()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(FastPS::start_waveform) ENABLED START -----*/
	
	//	Add your own code
	char strin[SHORT_LEN],strout[SHORT_LEN];
	memset(strin,0,SHORT_LEN);memset(strout,0,SHORT_LEN);

	sprintf(strin,"TRIG:?");
	write_cmd(strin,strout,txRxSleep);	
	//#MRI:2.3121
	if (strncmp(strout,"#TRIG:OFF",9) == 0)
		*attr_WaveformExternalTrigger_read = false;
	else
		*attr_WaveformExternalTrigger_read = true;

	memset(strin,0,SHORT_LEN);
	if (*attr_WaveformExternalTrigger_read)
		sprintf(strin,"WAVEFORM:TRIGGER:START");
	else
		sprintf(strin,"WAVEFORM:START");

	write_cmd(strin,strout,txRxSleep);
	
	/*----- PROTECTED REGION END -----*/	//	FastPS::start_waveform
}
//--------------------------------------------------------
/**
 *	Command Reset related method
 *	Description: Reset status errors
 *
 */
//--------------------------------------------------------
void FastPS::reset()
{
	DEBUG_STREAM << "FastPS::Reset()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(FastPS::reset) ENABLED START -----*/
	
	//	Add your own code
	char strin[SHORT_LEN],strout[SHORT_LEN];
	memset(strin,0,SHORT_LEN);memset(strout,0,SHORT_LEN);

	sprintf(strin,"MRESET");
	write_cmd(strin,strout,txRxSleep);
	
	/*----- PROTECTED REGION END -----*/	//	FastPS::reset
}
//--------------------------------------------------------
/**
 *	Command HwReset related method
 *	Description: Reboot the power supply
 *
 */
//--------------------------------------------------------
void FastPS::hw_reset()
{
	DEBUG_STREAM << "FastPS::HwReset()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(FastPS::hw_reset) ENABLED START -----*/
	
	//	Add your own code
	char strin[SHORT_LEN],strout[SHORT_LEN];
	memset(strin,0,SHORT_LEN);memset(strout,0,SHORT_LEN);

	sprintf(strin,"HWRESET");
	write_cmd(strin,strout,txRxSleep);
	
	/*----- PROTECTED REGION END -----*/	//	FastPS::hw_reset
}
//--------------------------------------------------------
/**
 *	Command StopWaveform related method
 *	Description: Start Waveform
 *
 */
//--------------------------------------------------------
void FastPS::stop_waveform()
{
	DEBUG_STREAM << "FastPS::StopWaveform()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(FastPS::stop_waveform) ENABLED START -----*/
	
	//	Add your own code
	char strin[SHORT_LEN],strout[SHORT_LEN];
	memset(strin,0,SHORT_LEN);memset(strout,0,SHORT_LEN);

	sprintf(strin,"WAVEFORM:STOP");
	write_cmd(strin,strout,txRxSleep);
	
	/*----- PROTECTED REGION END -----*/	//	FastPS::stop_waveform
}
//--------------------------------------------------------
/**
 *	Method      : FastPS::add_dynamic_commands()
 *	Description : Create the dynamic commands if any
 *                for specified device.
 */
//--------------------------------------------------------
void FastPS::add_dynamic_commands()
{
	/*----- PROTECTED REGION ID(FastPS::add_dynamic_commands) ENABLED START -----*/
	
	//	Add your own code to create and add dynamic commands if any
	
	/*----- PROTECTED REGION END -----*/	//	FastPS::add_dynamic_commands
}

/*----- PROTECTED REGION ID(FastPS::namespace_ending) ENABLED START -----*/

//	Additional Methods
//+------------------------------------------------------------------
/**
 *	method:	FastPS::write_cmd
 *
 */
//+------------------------------------------------------------------
long FastPS::write_cmd(char *strin, char *strout, int sleep_time) 
{
	int num;

	mutex->lock();
	sock_flush();

	/* add <cr> */
	strin[strlen(strin)]=13;
	strin[strlen(strin)+1]=0;	

	if ((num = write(sock, strin, strlen(strin))) < 0) {
		TangoSys_MemStream out_stream;
		out_stream << strerror(errno) << ends;	
		ERROR_STREAM << out_stream.str() << endl;
		mutex->unlock();	
		Tango::Except::throw_exception(
			(const char *) "Network communication error (writing)",
			out_stream.str(),
			(const char *) "FastPS::write_cmd", Tango::ERR);
	}
	INFO_STREAM << "FastPS::write_cmd: strin: " << strin << endl;

	omni_thread::sleep(0,sleep_time*1000000); // convert millisecond to nanosecond	

	if ((num = recv(sock, strout, LONG_LEN,0)) < 0) {
		TangoSys_MemStream out_stream;
		out_stream << strerror(errno) << ends;	
		ERROR_STREAM << out_stream.str() << endl;	
		mutex->unlock();		
		Tango::Except::throw_exception(
			(const char *) "Network communication error (reading)",
			out_stream.str(),
			(const char *) "FastPS::write_cmd", Tango::ERR);
	}
	INFO_STREAM << "FastPS::write_cmd: strout: " << strout << endl;

	if (num > 0) 
		gettimeofday(&socket_state_time,NULL);
	
	if (strncmp("#NAK:",strout,5) == 0) {
		TangoSys_MemStream out_stream;
		out_stream << error_mapping[atoi(strout+5)] << ends;	
		ERROR_STREAM << out_stream.str() << endl;
		mutex->unlock();
		Tango::Except::throw_exception(
			(const char *) "Network communication error (error message)",
			out_stream.str(),
			(const char *) "FastPS::write_cmd", Tango::ERR);
	}

	sock_flush();
	mutex->unlock();	
	return num;

}


//+------------------------------------------------------------------
/**
 *	method:	FastPS::sock_connect() 
 *
 */
//+------------------------------------------------------------------
int FastPS::sock_connect() 
{
	int connect_err_flag = 0,arg;
	TangoSys_MemStream out_stream;
	

	mysocket.sin_port = htons( tcpPort ); 
	mysocket.sin_addr.s_addr = inet_addr(ipAddress.c_str());
	mysocket.sin_family = AF_INET; 
	if (( sock = socket(AF_INET, SOCK_STREAM, 0)) < 0 ) {
		connect_err_flag = 1;
		cout << "Open device " << ipAddress << " failed, " << strerror(errno) << ends;			
		set_state(Tango::FAULT); 
		set_status(out_stream.str());	
		return connect_err_flag;
	}
	else
		connect_err_flag = 0;	
		
	/* getting socket parameters */
	if( (arg = fcntl(sock, F_GETFL, NULL)) < 0) {
		connect_err_flag = 1;
		cout << "Reading socket setting " << ipAddress << " failed, " << strerror(errno) << ends;		
		set_state(Tango::FAULT); 
		set_status(out_stream.str());
		return connect_err_flag;		
	}
	else
		connect_err_flag = 0;
			
	/* setting non blocking socket */						
	arg |= O_NONBLOCK;		
	if (fcntl (sock, F_SETFL, arg) < 0) {			
		connect_err_flag = 1;
		cout << "Setting socket " << ipAddress << " failed, " << strerror(errno) << ends;		
		set_state(Tango::FAULT); 
		set_status(out_stream.str());
		return connect_err_flag;		
	}
	else
		connect_err_flag = 0;	

	/* connect to device */		
	/* do not generate exception because non blocking socket configured */
	connect ( sock, (const sockaddr*)&mysocket, sizeof(mysocket));
	
	cout << "Connected to " << ipAddress << endl;		

	sock_flush();

	return connect_err_flag;

}	

//+------------------------------------------------------------------
/**
 *	method:	FastPS::sock_flush
 *
 */
//+------------------------------------------------------------------
void FastPS::sock_flush() 
{
	char trash[SHORT_LEN];	
	int count=0;
	while (recv(sock, trash, SHORT_LEN,0) >= 0) {
		count++;
		if (count > 10)
			break;
	};	/* flush socket */
}

//+------------------------------------------------------------------
/**
 *	method:	FastPS::set_DevState
 *
 */
//+------------------------------------------------------------------
void FastPS::set_DevState(void) 
{
	char strin[SHORT_LEN],strout[SHORT_LEN];
	int32_t state_mask;
	stringstream s;	

	memset(strin,0,SHORT_LEN);memset(strout,0,SHORT_LEN);
	sprintf(strin,"MST:?");
	write_cmd(strin,strout,txRxSleep);	
	state_mask = (int32_t)strtol(strout+strlen(strin)-1, NULL, 16);

	if (state_mask & (1 << FASTPS_ON)) {
		s << "PS ON" << endl;
		set_state(Tango::ON);
	}
	else {
		s << "PS OFF" << endl;
		set_state(Tango::OFF);
	}

	if (state_mask & (1 << FASTPS_LOCAL)) {
		s << "Local control" << endl;
	}

	if (state_mask & (1 << FASTPS_CTRL)) {
		*attr_VoltageControl_read = true;
		*attr_CurrentControl_read = false;
	}
	else {
		*attr_VoltageControl_read = false;
		*attr_CurrentControl_read = true;
	}

	if (state_mask & (1 << FASTPS_EXTREF)) 
		*attr_AnalogControl_read = true;
	else
		*attr_AnalogControl_read = false;

	if (state_mask & (1 << FASTPS_RAMPING)) {
		s << "PS ramping" << endl;
		set_state(Tango::MOVING);
	}

	if (state_mask & (1 << FASTPS_WAVE)) {
		s << "PS running waveform" << endl;
		set_state(Tango::RUNNING);
	}

	if (state_mask & (1 << FASTPS_FAULT)) {
		s << "PS FAULT" << endl;
		set_state(Tango::FAULT);
	}


	if (state_mask & (1 << FASTPS_IOVCERR )) {
		s << "input overcurrent error" << endl;
	}

	if (state_mask & (1 << FASTPS_CROWBARERR)) {
		s << "Crowbar protection intervention" << endl;
	}

	if (state_mask & (1 << FASTPS_OVTERR)) {
		s << "Over temperature" << endl;
	}

	if (state_mask & (1 << FASTPS_DCLINKERR)) {
		s << "DC-link undervoltage" << endl;
	}

	if (state_mask & (1 << FASTPS_EARTHLEAKERR)) {
		s << "earth current leakage fault" << endl;
	}

	if (state_mask & (1 << FASTPS_EARTHFUSEERR)) {
		s << "Earth fuse is blown" << endl;
	}

	if (state_mask & (1 << FASTPS_REGERR)) {
		s << "Regulation fault" << endl;
	}

	if (state_mask & (1 << FASTPS_RIPPLEERR)) {
		s << "Excessive ripple" << endl;
	}

	if (state_mask & (1 << FASTPS_INT1ERR)) {
		s << "Interlock 1 tripped" << endl;
	}

	if (state_mask & (1 << FASTPS_INT2ERR)) {
		s << "Interlock 2 tripped" << endl;
	}

	if (state_mask & (1 << FASTPS_OVPERR)) {
		s << "Over power" << endl;
	}

	set_status(s.str());
}

//+------------------------------------------------------------------
/**
 *	method:	FastPS::read_waveform
 *
 */
//+------------------------------------------------------------------
void FastPS::read_waveform(double *waveform, int32_t *size) 
{
	char strin[LONG_LEN],strout[LONG_LEN];
	int32_t cnt = 0;

	memset(strin, 0, LONG_LEN);memset(strout, 0, LONG_LEN);
	sprintf(strin,"WAVEFORMS:POINTS:?");
	write_cmd(strin,strout,txRxSleep*10);
	char *token = strtok(strout, ":");

	while (token != NULL) {
		waveform[cnt++] = atof(token);
		token = strtok(NULL,":");
	}
	*size = cnt;
	
}

//+------------------------------------------------------------------
/**
 *	method:	FastPS::set_waveform
 *
 */
//+------------------------------------------------------------------
void FastPS::write_waveform(double *waveform, int32_t size, int32_t time_scale) 
{
	char strin[LONG_LEN],strout[LONG_LEN];
	int32_t cnt = 0;

	memset(strin, 0, LONG_LEN);memset(strout, 0, LONG_LEN);
	sprintf(strin,"WAVEFORMS:POINTS");
	for (int i = 0; i < size; i++) {
		for (int j = 0; j < time_scale; j++) {
				sprintf(strin+strlen(strin),":%.5f",waveform[i]);
				cnt++;
				if (cnt > FASTPS_MAX_WAVE)
					break;
		}
	}
	if (cnt < FASTPS_MIN_WAVE) {
		for (int i = cnt; i < FASTPS_MIN_WAVE; i++) {
			sprintf(strin+strlen(strin),":%.5f",waveform[size-1]);
		} 
	}
	write_cmd(strin,strout,txRxSleep*10);

}



/*----- PROTECTED REGION END -----*/	//	FastPS::namespace_ending
} //	namespace
